"""
GITLAB Upstream CI environment
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from pykrete.args import CiIo


class GitLabUpstreamEnv:
    """GitLab upstream environment"""

    @property
    def project_id(self):
        """
        :return: Upstream project ID
        """
        return self._project_id

    @property
    def pipeline_id(self):
        """
        :return: Upstream pipeline ID
        """
        return self._pipeline_id

    @property
    def build_job_prefix(self):
        """
        :return: Upstream build job prefix
        """
        return self._build_job_prefix

    @property
    def key(self):
        """
        :return: Upstream token key
        """
        return self._key

    def __init__(self, ci_io=None):
        """Initialize this instance

        :param ci_io: CI environment's IO manager (optional, defaults to pykrete.args.CiIo)
        """
        self.__ci_io = ci_io if ci_io else CiIo()
        (self._project_id, self._pipeline_id, self._build_job_prefix, self._key) =\
            self.__get_upstream_parts()

    def __get_upstream_parts(self):
        return tuple(
            [self.__get_upstream(part) for part in
             ['project ID', 'pipeline ID', 'build job prefix', 'key']])

    def __get_upstream(self, part):
        return self.__ci_io.read_env(f'upstream {part}')



"""
GIT repository push management
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging
from git import GitCommandError
from pykrete.credentials import SshTarget
from .git_urls import GitUrls
from .git_ssh import GitSsh


class GitPush(SshTarget):
    """Handles pushing to a git repo"""

    def __init__(self, repo, credentials=None, remote='origin'):
        """Initialize this repository

        :param repo: git repository
        :param credentials: git credential (optional, defaults to None)
        """
        self._logger = logging.getLogger(__name__)
        self._repo = repo
        self._credentials = credentials
        self._remote = repo.remotes[remote]
        self._urls = GitUrls(self._remote)
        self._active_credentials = None

    def __enter__(self):
        if self._credentials:
            self.accept(self._credentials)
        return self

    def __exit__(self, *args):
        if self._active_credentials:
            self._active_credentials.unset(*args)

    def push(self, item):
        """Push the item

        :param item: item to push
        """
        try:
            self._logger.debug('Pushing %s', item)
            self._remote.push(item)
            self._logger.debug('%s pushed', item)
        except GitCommandError as exc:
            self._logger.error('Push failed: %s', exc)
            raise

    def set_ssh(self, private_key_file_path, ssh_port):
        """Set SSH private key to the specified path

        :param private_key_file_path: Path to private key file
        :param ssh_port: SSH port
        """
        self._active_credentials = GitSsh(
            self._repo.git.custom_environment, self._urls, private_key_file_path, ssh_port)



"""
Pykrete repo testbed
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging


class GitUrlsTestbed:
    """simulated"""

    _logger = logging.getLogger(__name__)

    @property
    def configured_ssh_urls(self):
        """simulated"""
        return []

    def __init__(self):
        """simulated"""
        self.replaced_urls = []
        self.ssh_replaced = False

    def configure_ssh_url_and_save_original(self, ssh_port):
        """simulated"""
        self._logger.debug('configure_ssh_url_and_save_original called with port=%s', ssh_port)
        self.ssh_replaced = True
        return 'http', 'ssh'

    def replace_url(self, new, old):
        """simulated"""
        self._logger.debug('replace_url called, new=%s, old=%s', new, old)
        self.replaced_urls.append((new, old))


class WithTestbed:
    """simulated"""

    def __init__(self):
        """simulated"""
        self.entered = False
        self.exited = False
        self.args = None

    def __enter__(self):
        """simulated"""
        self.entered = True

    def __exit__(self, *args):
        """simulated"""
        self.exited = True
        self.args = args


class GitRepoRemoteTestbed:
    """simulated"""

    def __init__(self, urls):
        """simulated"""
        self.urls = urls
        self.urls_set = []

    def set_url(self, new, old):
        """simulated"""
        self.urls_set.append((new, old))
        self.urls[self.urls.index(old)] = new



"""
Feature revision types
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from . import Version


class Formatting:
    """Version formatter"""
    @property
    def for_python(self):
        """Create python version string for the formatted version

        :return: The version string
        """
        return ''.join(
            [
                str(self._version.major),
                '.',
                str(self._version.minor),
                ['pa', 'a', 'b', 'rc', '.'][self._version.revision_type.value],
                str(self._version.build),
            ])

    def __init__(self, version: Version):
        """Initialize this instance to format the specified version

        :param version: Version to be formatted
        """
        self._version = version



"""
GIT tag-based version information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging
import re
from pykrete.args import CiIo
from pykrete.repo import Git
from .revision_type import RevisionType
from .version import Version
from .ci_revision_and_build import CiRevisionAndBuild
from .tag_version_bumper import TagVersionBumper


class TagVersion(Version):
    """Handle version in CI environment
    A version is constructed from the last version tag (tag name format '<major>.<minor>.<build>'
     indicating a release version, with a message format of 'ci_base_build <build>' indicating the
     CI build number for the first release with this major-minor combo). In the case of a release,
     a commit message containing '#major' will increment the major version, otherwise a commit
     message containing '#minor' will increment the minor version.
    Build is calculated as the current CI project-build# variable, minus the build specified in the
     last version tag's message if this isn't a release, or 0 if it is a release.
    Revision is read from the CI environment variables CI_COMMIT_REF_NAME, CI_MERGE_REQUEST_TITLE
     and CI_JOB_NAME:
        Release (4) - master branch on a job who'se name doesn't contain '_rc_'
        RC (3) - non-WIP merge request, master branch on a job who'se name contains '_rc_'
        Beta (2) - merge request with 'WIP' in the title
        Alpha (1) - none of the above
    """

    _empty_version = [0, 0, 0], 0

    def __init__(self, repo=Git(), ci_io=None):
        """Initializes this instance from CI environment

        :param repo: project's repository (optional, defaults to git for the current folder)
        :param ci_io: CI environment's IO manager (optional, defaults to pykrete.args.CiIo)
        """
        self._logger = logging.getLogger(__name__)
        self.__revision = CiRevisionAndBuild(ci_io if ci_io else CiIo())
        self.__repo = repo
        self.__tags = self.__read_tags()
        # pylint: disable=unbalanced-tuple-unpacking
        (major, minor, build) = self.__read_version_from_tags()
        super().__init__(major=major,
                         minor=minor,
                         revision=self.__revision.revision_type.value,
                         build=build,
                         revision_type=self.__revision.revision_type)

    def apply(self):
        """Apply this version to the relevant environment
        """
        if self.__revision.revision_type != RevisionType.Release:
            raise ValueError('Cannot apply non-release version')
        self.__repo.add_tag(f'{self.major}.{self.minor}.{self.build}',
                            f'ci_base_build {self.__revision.build}')

    def __read_tags(self):
        return self.__repo.get_tags_from(pattern=r'(\d+\.){2}\d+', is_must=False)

    def __read_version_from_tags(self):
        last_version_tag = self.__tags[0] if self.__tags else None
        version_parts, base_build =\
            self.__get_version_parts_and_base_build_from(last_version_tag)
        version_parts = self.__advance_version(version_parts, base_build)
        return tuple(version_parts)

    def __get_version_parts_and_base_build_from(self, tag):
        return self.__make_version_parts_and_base_build_from(tag) if tag else self._empty_version

    def __make_version_parts_and_base_build_from(self, tag):
        version_parts = [int(part) for part in tag[0].split('.')]
        message_build = re.findall(r'ci_base_build (\d+)', tag[1]) if tag[1] else None
        base_build = int(message_build[0]) if message_build else 0
        self._logger.debug('Read version parts: %s [base build %s]', version_parts, base_build)
        return version_parts, base_build

    def __advance_version(self, version_parts, base_build):
        return TagVersionBumper(self.__repo, self.__revision, base_build)\
            .advance_version(version_parts)



"""
GIT repository SSH management
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging


class GitSsh:
    """Handles pushing to a git repo"""

    def __init__(self, environment_setter, urls, private_key_file_path, ssh_port):
        """Set SSH private key to the specified path

        :param environment_setter: sets a git repository's custom environment
        :param urls: branch url management
        :param private_key_file_path: Path to private key file
        :param ssh_port: SSH port (optional, defaults to 22)
        """
        self._logger = logging.getLogger(__name__)
        self._urls = urls
        self._path = private_key_file_path
        self._environment = self._set_environment(environment_setter, private_key_file_path)
        if not self._urls.configured_ssh_urls:
            self._original_urls = self._urls.configure_ssh_url_and_save_original(ssh_port)

    def unset(self, *args):
        """Unset the credentials

        :param args: exit arguments
        """
        self._logger.debug('Disabling GIT SSH')
        self._environment.__exit__(*args)
        if self._original_urls:
            http_url, old_url = self._original_urls
            self._urls.replace_url(http_url, old_url)

    def _set_environment(self, environment_setter, private_key_file_path):
        ssh_cmd = 'ssh -o StrictHostKeyChecking=no -i ' \
                  + private_key_file_path.replace('\\', '\\\\')
        self._logger.debug('Enabling GIT SSH with command [%s]', ssh_cmd)
        environment = environment_setter(GIT_SSH_COMMAND=ssh_cmd)
        environment.__enter__()
        return environment



"""
Build manager base class
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging
from abc import abstractmethod
from pykrete.args import environ
from pykrete.versioning import Version, RevisionType


class BuildManager:
    """Manages building a python package"""

    _logger = logging.getLogger(__name__)

    def __init__(self, version: Version):
        """Initializes this instance
        :param version: Build version
        """
        self._version = version
        self._project = environ('CI_TARGET', 'the name of the project under src')

    def run(self):
        """Runs the build process"""
        self._clean_distribution()
        self._set_package_version()
        self._build()
        self._upload(self._version.revision_type == RevisionType.Release)

    @abstractmethod
    def _clean_distribution(self):
        """Removes previous distribution files"""

    @abstractmethod
    def _set_package_version(self):
        """Sets the package version in the project's version file"""

    @abstractmethod
    def _build(self):
        """Performs the actual build using build.py"""

    @abstractmethod
    def _upload(self, is_release):
        """Uploads the package

        :param is_release: flag this package as release
        """

    def __str__(self):
        """Gets a string representation of this object"""
        return f'Builder for version {self._version} of project "{self._project}"'



"""
Fatal issue handler
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging


def fatal(message, exception_class=None):
    """Log and raise a fatal error

    :param message: error message
    :param exception_class: Exception class (optional, use Exception if not supplied)
    :exception: always raises an exception of the specified type with the specified message
    """
    logging.getLogger(__name__).error(message)
    raise (exception_class if exception_class else Exception)(message)



"""
Python package
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging
from pykrete.versioning import Formatting, TagVersion, VersionPyVersion


class PythonPackage:
    """Python package"""

    @property
    def project(self):
        """
        :return: (string) name of project package under 'src'.
        """
        return self._project

    @property
    def version(self):
        """
        :return: (string) The project's version
        """
        return self._version

    def __init__(self, project):
        """Initializes this instance to analyze the specified project

        :param project: name of project package under 'src'
        """
        self._logger = logging.getLogger(__name__)
        self._project = project
        version_makers = self._collect_version_makers()
        self._version = self._get_version(version_makers)

    @property
    def long_description(self):
        """Gets the contents of the README.md file

        :return: Long description
        """
        with open('README.md', 'r') as readme:
            return readme.read()

    def _collect_version_makers(self):
        return [self._get_version_from_tag, self._get_version_from_version_py]

    @staticmethod
    def _get_version(version_makers):
        """Gets the package version

        :return: The version
        :exception IOError: Version read failed
        """
        version = None
        for maker in version_makers:
            version = maker()
            if version:
                break
        if not version:
            raise IOError('Unable to get version')

        return Formatting(version).for_python

    def _get_version_from_tag(self):
        try:
            version = TagVersion()
            self._logger.debug('Got version from tag: %s', version)
            return version
        except IOError as exc:
            self._logger.debug('Unable to get version from tag [%s]', exc)
            return None

    def _get_version_from_version_py(self):
        try:
            version = VersionPyVersion(self._project)
            self._logger.debug('Got version from version.py: %s', version)
            return version
        except IOError as exc:
            self._logger.debug('Unable to get version from version.py of %s [%s]',
                               self._project, exc)
            return None

    def __str__(self):
        return f'{self._project} v{self._version}'



"""
Pykrete Args tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging
import unittest
import argparse
from pykrete.args import build_version, exiting_file, yes_or_no, environ, CiIo


class PykreteArgsTestCase(unittest.TestCase):
    """Unit tests for pykrete's args module
    """

    _logger = logging.getLogger(__name__)

    def test_build_version_good(self):
        """Verifies correct parsing of a legal version
        """
        source = '1.2.3.4'
        version = build_version(source)
        self.assertEqual(version.string, source, 'read version differs from source')
        self._logger.debug(':'.join(version.groups()[1:4]))

    def test_build_version_short(self):
        """Verifies rejection of a short version
        """
        with self.assertRaises(argparse.ArgumentTypeError):
            self._logger.debug(build_version('1.2.3'))

    def test_build_version_bad(self):
        """Verifies rejection of a non-version
        """
        with self.assertRaises(argparse.ArgumentTypeError):
            self._logger.debug(build_version('Blue. No, yellOOOOOOOW!!'))

    def test_existing_file_good(self):
        """Verifies acceptance of an existing file
        """
        source = 'README.md'
        self.assertEqual(exiting_file(source), source, 'read file name differs from source')

    def test_existing_file_bad(self):
        """Verifies rejection of a non-existing file
        """
        with self.assertRaises(argparse.ArgumentTypeError):
            self._logger.debug(exiting_file('no_such_file.txt'))

    def test_yes_no_good(self):
        """Verifies correct parsing of yes/no answers
        """
        cases = {True: ['yes', 'Yes', 'YES'], False: ['no', 'No', 'NO']}
        for key, sources in cases.items():
            for source in sources:
                self.assertEqual(key, yes_or_no(source), f'{source} not {key}')

    def test_yes_no_bad(self):
        """Verifies rejection of a non-yes/no answer
        """
        with self.assertRaises(argparse.ArgumentTypeError):
            self._logger.debug(yes_or_no('Huh? I don\'t know that! EEEEEEAAAARRGH!!!'))

    def test_environ_existing(self):
        """Verifies reading existing environment variable
        """
        self._logger.debug('Read path: %s', environ('PATH', 'search path'))

    def test_environ_non_existing_without_role(self):
        """Verifies reading allowed non-existing environment variable
        """
        self._logger.debug('Read non-existing: %s', environ('NO_IM_NOT'))

    def test_environ_non_existing(self):
        """Verifies attempting to read not-allowed non-existing environment variable raises an error
        """
        with self.assertRaises(IOError):
            self._logger.debug('Read non-existing: %s',
                               environ('ITS_JUST_A_FLESH_WOUND',
                                       'Look, you stupid b***ard, you''ve got no arms left!'))

    def test_ciio_existing(self):
        """Verifies reading existing environment variable
        """
        part = 'part1'
        target = CiIo(ci_spec={part: 'PATH'})
        self._logger.debug('Read path: %s', target.read_env(part))

    def test_ciio_non_existing_with_non_false_default(self):
        """Verifies reading default non-False non-existing environment variable
        """
        part = 'part2'
        target = CiIo(ci_spec={part: 'OH_YES_IT_IS'})
        self._logger.debug('Read non-existing: %s', target.read_env(part, 'Oh no it isn''t'))

    def test_ciio_non_existing_with_false_default(self):
        """Verifies reading default False non-existing environment variable
        """
        part = 'part3'
        target = CiIo(ci_spec={part: 'OH_YES_IT_IS'})
        self._logger.debug('Read non-existing: %s', target.read_env(part, False))

    def test_ciio_non_existing_without_default(self):
        """Verifies attempting to read no-default non-existing environment variable raises an error
        """
        part = 'part4'
        target = CiIo(ci_spec={part: 'OH_YES_IT_IS'})
        with self.assertRaises(IOError):
            self._logger.debug('Read non-existing: %s', target.read_env(part))


if __name__ == '__main__':
    unittest.main()



"""
CI environment-based version information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from pykrete.args import CiIo
from .version import Version
from .ci_revision_and_build import CiRevisionAndBuild


class CiVersion(Version):
    """Handle version in CI environment
    Version parts are read from CI variables CI_VERSION_MAJOR, CI_VERSION_MINOR and CI_PIPELINE_IID
    Revision is read from CI_COMMIT_REF_NAME, CI_MERGE_REQUEST_TITLE and CI_JOB_NAME:
        Release - in master branch and job's name doesn't contain '_rc_'
        RC - in non-WIP merge request, or in master branch with a job who'se name contains '_rc_'
        Beta - in merge request with 'WIP' in the title
        Alpha - none of the above
    """

    def __init__(self, ci_io=None):
        """Initializes this instance from CI environment

        :param ci_io: CI environment's IO manager (optional, defaults to pykrete.args.CiIo)
        """
        self.__ci_io = ci_io if ci_io else CiIo()
        revision = CiRevisionAndBuild(self.__ci_io)
        major, minor = self._read_parts('major', 'minor')
        super().__init__(major=major,
                         minor=minor,
                         revision=revision.revision_type.value,
                         build=revision.build,
                         revision_type=revision.revision_type)

    def apply(self):
        """Not implemented yet"""
        raise NotImplementedError('No permanent CI change implemented yet')

    def _read_parts(self, *args):
        return tuple([int(self.__ci_io.read_env(f'{part} version')) for part in args])



"""
Pykrete CI features tests base
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging
import unittest


class PykreteCiEchoTestCase(unittest.TestCase):
    """Unit tests base for pykrete's CI dependent features
    """

    _logger = logging.getLogger(__name__)

    def _echo(self, what, who):
        self._logger.debug('CiVersion is reading \'%s\' and will get \'%s\'', who, what)
        return None if what == 'None' else what



"""
GIT repository URL management
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import re
import logging


class GitUrls:
    """Handles the URLs attached to a remote"""

    __ssh_prefix = 'ssh://git@'

    @property
    def configured_ssh_urls(self):
        """
        :return SSH urls configured for this remote
        """
        return [url for url in
                self._remote.urls if url.startswith(self.__ssh_prefix)]

    def __init__(self, remote):
        """Initialize this instance

        :param remote: git remote
        """
        self._logger = logging.getLogger(__name__)
        self._remote = remote

    def configure_ssh_url_and_save_original(self, ssh_port):
        """Configures an SSH url instead of existing HTTP url

        :param ssh_port: SSH port
        :return: (original, new) URLs
        """
        http_url, server, project = self._get_http_url_server_and_project_for_remote()
        ssh_url = self._make_ssh_url(server, project, ssh_port)
        self.replace_url(ssh_url, http_url)
        return http_url, ssh_url

    def replace_url(self, new, old):
        """Replace remote URL

        :param new: new URL to use
        :param old: old URL to replace
        """
        self._logger.debug('replacing URLS, new=%s, old=%s', new, old)
        self._remote.set_url(new, old)

    def _get_http_url_server_and_project_for_remote(self):
        pattern = re.compile(r'http.?://.*?([^:@/]+)/(.*)')
        matches = [(url, ) + tuple(match.groups()) for url, match in
                   [(url, pattern.match(url)) for url in self._remote.urls]
                   if match]
        if not matches:
            raise IOError(f'No HTTP* URLs found in {self._remote}')
        return matches[0]

    def _make_ssh_url(self, server, project, port):
        port_spec = '' if str(port) == '22' else f':{port}'
        return f'{self.__ssh_prefix}{server}{port_spec}/{project}'



"""
distutils pylint project-test command
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from sys import argv
from .pylint_command import PylintCommand


class SelfTestCommand(PylintCommand):
    """A custom command to run static-test on build.py"""

    def finalize_options(self):
        """Post-process received options"""
        assert 'bypass', 'project name not specified'

    def _get_target_python_files_and_folders(self):
        """provide called script and built.py as targets

        :return: Called script and build.py
        """
        return [argv[0], 'build.py']



"""
distutils command for building local GIT project
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from distutils.log import INFO
from distutils.cmd import Command
from pykrete.versioning import TagVersion
from pykrete.builder import PythonBuildManager
from pykrete.repo import Git
from pykrete.credentials import CiSshPrivateKey


class BuildGitSshTag(Command):
    """A custom command for building local GIT project where SSH is used for pushing,
     and the version is tag-based
    """
    description = 'Build the current GIT project'
    user_options = []

    def initialize_options(self):
        """Set default values for options"""

    def finalize_options(self):
        """Post-process received options"""

    def run(self):
        """Performs the build"""
        with CiSshPrivateKey() as key:
            repo = Git(credentials=key)
            version = TagVersion(repo=repo)
            manager = PythonBuildManager(version)
            self.announce(str(manager), level=INFO)
            manager.run()
        self.announce('Build succeeded', level=INFO)



"""
Machine's states
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from abc import abstractmethod

from .parsing_state import ParsingState


class BetweenArguments(ParsingState):
    """Transitions:
    quote -> start quoted argument
    whitespace -> keep looking
    all other -> start unquoted argument
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, context):
        super().__init__(context)
        self.__transitions_from_chars_to_new_states =\
            self.__make_transitions_from_whitespace_and_quote_chars_to_new_states()

    def _get_default_next_state(self):
        return InAnUnquotedArgument

    def _get_dict_of_chars_to_new_states(self):
        return self.__transitions_from_chars_to_new_states

    def __make_transitions_from_whitespace_and_quote_chars_to_new_states(self):
        transitions_from_chars_to_new_states =\
            self.__make_transitions_from_whitespace_chars_to_same_state()
        transitions_from_chars_to_new_states.update(
            self.__make_transitions_from_quote_chars_to_quoted_arg())
        return transitions_from_chars_to_new_states

    def __make_transitions_from_quote_chars_to_quoted_arg(self):
        return {char: InAQuotedArgument for char in self._quote_chars}

    def __make_transitions_from_whitespace_chars_to_same_state(self):
        return {char: None for char in self._whitespace_chars}


class _InAnArgument(ParsingState):
    """Transitions:
    ending character -> argument ends
    all other -> argument continues
    """
    def __init__(self, context):
        super().__init__(context)
        self._context.flush_current_arg()
        self.__chars_to_new_states =\
            self.__make_transitions_from_non_argument_chars_to_between_arguments_state()

    def _on_new_char(self, char):
        super()._on_new_char(char)
        self._context.append_last_char_to_current_arg()

    def _get_default_next_state(self):
        return None

    def _get_dict_of_chars_to_new_states(self):
        return self.__chars_to_new_states

    def __make_transitions_from_non_argument_chars_to_between_arguments_state(self):
        return {char: BetweenArguments for char in self._get_non_argument_chars()}

    @abstractmethod
    def _get_non_argument_chars(self):
        """
        :return: a list of characters which mark the and of the argument
        """


class InAnUnquotedArgument(_InAnArgument):
    """Don't quote me on that"""
    def __init__(self, context):
        super().__init__(context)
        context.append_last_char_to_current_arg()

    def _get_non_argument_chars(self):
        return self._whitespace_chars


class InAQuotedArgument(_InAnArgument):
    """And you can quote me on that"""
    def __init__(self, context):
        self.__quote_char = [context.last_char]
        super().__init__(context)

    def _get_non_argument_chars(self):
        return self.__quote_char



"""
Version information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from abc import abstractmethod


class Version:
    """Represents a package version
    """

    @property
    def major(self):
        """
        :return: Major part of the version
        """
        return self._major

    @property
    def minor(self):
        """
        :return: Minor part of the version
        """
        return self._minor

    @property
    def revision(self):
        """
        :return: Revision part of the version
        """
        return self._revision

    @property
    def build(self):
        """
        :return: Build part of the version
        """
        return self._build

    @property
    def revision_type(self):
        """
        :return: Revision type of the version
        """
        return self._revision_type

    def __init__(self, *args, **kwargs):
        """Initializes this instance with the specified values

        :param args: one of:
            - a single tuple of (major, minor, revision, build, revision_type)
            - these five values individually
            - a Version object
        :param kwargs: specify values with labels: major, minor, revision, build, revision_type
        """
        (major, minor, revision, build, revision_type) = \
            Version.__get_version_parts_from_args(args) if args \
            else Version.__get_version_parts_from_kwargs(kwargs)
        self._major = major
        self._minor = minor
        self._revision = revision
        self._build = build
        self._revision_type = revision_type

    @abstractmethod
    def apply(self):
        """Apply this version to the relevant environment
        """

    def __str__(self):
        return f'{self.major}.{self.minor}.{self.revision}.{self.build} {self.revision_type.name}'

    def __ge__(self, other):
        return not self.__lt__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __lt__(self, other):
        for comparison in self.__make_comparisons(other):
            if comparison[0] < comparison[1]:
                return True
            if comparison[0] > comparison[1]:
                return False
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        for comparison in self.__make_comparisons(other):
            if comparison[0] != comparison[1]:
                return False
        return self.revision_type == other.revision_type

    def __make_comparisons(self, other):
        comparisons = [(self.major, other.major), (self.minor, other.minor),
                       (self.revision, other.revision), (self.build, other.build)]
        return comparisons

    @staticmethod
    def __get_version_parts_from_kwargs(kwargs):
        return (kwargs['major'], kwargs['minor'], kwargs['revision'], kwargs['build'],
                kwargs['revision_type'])

    @staticmethod
    def __get_version_parts_from_args(args):
        first_arg = args[0]
        first_arg_type = type(first_arg)
        if first_arg_type == tuple:
            return first_arg
        if issubclass(first_arg_type, Version):
            return (first_arg.major, first_arg.minor, first_arg.revision, first_arg.build,
                    first_arg.revision_type)
        return args



"""
distutils pylint-command base
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from re import findall
from abc import abstractmethod
from distutils.errors import DistutilsModuleError
from distutils.log import INFO, ERROR, DEBUG
from ..project_command import ProjectCommand
from ...calls import CheckedCall


class PylintCommand(ProjectCommand):
    """A custom command to run static-test on python files"""

    description = 'Run pylint static analysis on a package'

    def run(self):
        self.__use_pylint_to_verify_each(self._get_target_python_files_and_folders())

    @abstractmethod
    def _get_target_python_files_and_folders(self):
        """
        :return: a list of paths to python code files/folders
        """

    def __use_pylint_to_verify_each(self, targets):
        failures = []
        for target in targets:
            cmd = self.__pylint_test_target(target)
            if not cmd.success:
                failures.append(target)
            else:
                self.__announce_target_rating(cmd.get_output())
        if failures:
            raise DistutilsModuleError(f'Static analysis failed on {failures}')

    def __announce_target_rating(self, output):
        self.announce('Rating: ' + '<--'.join(findall(r'\d+\.\d+/\d+', output)), INFO)

    def __pylint_test_target(self, target):
        self.announce(f'Static analysis of: {target}', level=INFO)
        cmd = CheckedCall(f'python -m pylint {target}')
        self.announce(cmd.get_output(), DEBUG if cmd.success else ERROR)
        return cmd



"""
Repository change aggregation
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""


class Change:
    """Used for collecting and reviewing repository changes"""
    @property
    def log(self):
        """
        :return: change log (titles)
        """
        return self.__join_str_list(self._log)

    @property
    def details(self):
        """
        :return: change details (descriptive messages)
        """
        return self.__join_str_list(self._details)

    def __init__(self):
        """Initializes this instance to an empty change"""
        self._log = []
        self._details = []

    def append_log(self, log):
        """Appends to the change log

        :param log: log to append (can be a string, list or tuple of strings)
        """
        self.__flatten_convert_and_append_to_list(self._log, log, str)

    def append_details(self, details):
        """Appends to the change details

        :param details: log to append (can be a string, list or tuple of strings)
        """
        self.__flatten_convert_and_append_to_list(self._details, details, str)

    def __str__(self):
        """returns a string representation of this change"""
        return '\n====\n'.join([
            f'Log:\n{self.log}',
            f'Details:\n{self.details}',
        ])

    @staticmethod
    def __join_str_list(str_list):
        return '\n'.join(str_list)

    @staticmethod
    def __flatten_convert_and_append_to_list(a_list, item, converter):
        items = item if isinstance(item, (list, tuple)) else [item]
        a_list.extend([converter(part) for part in items])



"""
Pattern and common classes
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""


def write_lines(file, lines):
    """Write lines to file

    :param file: target file
    :param lines: lines to write
    """
    file.writelines('\n'.join(lines) + '\n')



"""
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging
from subprocess import Popen, PIPE

from ._parse_command_into_call_vector import parse_command_into_call_vector


class CheckedCall:
    """Used to execute shell commands and hold the execution results"""

    @property
    def stdout(self):
        """
        :return: (string) command's standard output
        """
        return self._stdout

    @property
    def stderr(self):
        """
        :return: (string) command's standard error
        """
        return self._stderr

    @property
    def success(self):
        """
        :return: (bool) True if exit-code was 0, False otherwise
        """
        return self._success

    def __init__(self, cmd):
        """Initializes this instance with the specified command

        :param cmd: (string/list) shell command
        """
        logger = logging.getLogger(__name__)
        logger.debug('Running command: %s', cmd)
        command_vector = self.__get_vector_from_command(cmd)
        self._process = Popen(command_vector, shell=True, stdout=PIPE, stderr=PIPE)
        (self._stdout, self._stderr) = self._run_and_get_streams()
        self._success = self._process.returncode == 0
        logger.debug('Command result:\n--- >> start output << ---\n%s\n--- >> end output << ---',
                     self.get_output())

    def get_output(self):
        """Gets the command's output

        :return: stdout + stderr
        """
        return f'{self._stdout}\n{self._stderr}'

    def __str__(self):
        return f'{self._process.args}\n{self.get_output()}'

    def _run_and_get_streams(self):
        """
        :return (string, string): (stdout, stderr) strings
        """
        return tuple([stream.decode('ascii') for stream in self._process.communicate()])

    @staticmethod
    def __get_vector_from_command(cmd):
        return cmd if isinstance(cmd, list) else parse_command_into_call_vector(str(cmd))



"""
Visitor pattern base class
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from abc import ABCMeta, abstractmethod


class Visitor(metaclass=ABCMeta):
    """Visitor pattern base class"""

    @abstractmethod
    def visit(self, target):
        """Visit this instance

        :param target: visited target
        """



"""
Packages management
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

from .python_package import PythonPackage
from .install_python_packages import install_python_packages

__all__ = ['PythonPackage', 'install_python_packages']



"""
Pykrete module build script
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from setuptools import setup, find_packages

__version__ = '0.99.99'
exec(open('./src/pykrete/version.py').read())
print('Installing pykrete v.' + __version__)
with open('./requirements.txt') as f:
    REQUIREMENTS = f.read().splitlines()
with open('./README.md') as f:
    LONG_DESCRIPTION = f.read()
setup(
    name='pykrete',
    version=__version__,
    license='MIT',
    author='Shai A. Bennathan',
    author_email='shai.bennathan@intel.com',
    description='Build script foundation',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='http://ait-tool-center.iil.intel.com/',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=REQUIREMENTS
)



"""
Pykrete versioning.tag_version tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import unittest
import logging
from pykrete.args import CiIo
from pykrete.versioning import TagVersion, RevisionType, Version
from pykrete.repo import Change
from .versioning import PykreteVersioningTestCase
from .repo_testbed import RepoTestbed


class PykreteVersioningTagVersionTestCase(PykreteVersioningTestCase):
    """Unit tests for pykrete's versioning module's TagVersion class
    """

    _logger = logging.getLogger(__name__)
    _local_alpha = {
        'build version': '3',
        'branch name': 'custom',
        'merge request title': 'None'}
    _local_beta = {
        'build version': '111',
        'branch name': 'custom',
        'merge request title': 'WIP: something'}
    _change_request_rc = {
        'build version': '1234',
        'branch name': 'custom',
        'merge request title': 'something'}
    _master_rc = {
        'build version': '1234',
        'branch name': 'master',
        'merge request title': 'something',
        'job name': 'some_rc_job'}

    def test_tag_version_local_alpha_no_tags(self):
        """Verifies handling of no simulated tags present"""
        self._assert_tag_version_local(
            spec=self._local_alpha,
            tags=[None],
            change_details="",
            expected_version=Version((0, 0, 1, 3, RevisionType.Alpha)))

    def test_tag_version_local_alpha_first_minor_bump(self):
        """Verifies handling of simulated minor bump without a previous release on non-release"""
        self._assert_tag_version_local(
            spec=self._local_alpha,
            tags=[None],
            change_details="this won't cause a #minor bump",
            expected_version=Version((0, 0, 1, 3, RevisionType.Alpha)))

    def test_tag_version_local_release_first_minor_bump(self):
        """Verifies handling of simulated minor bump without a previous release on release"""
        self._assert_tag_version_local(
            spec=self._make_release_spec_for_build('4'),
            tags=[None],
            change_details="this will cause a #minor bump",
            expected_version=Version((0, 1, 4, 0, RevisionType.Release)))

    def test_tag_version_local_beta_major_bump(self):
        """Verifies handling of simulated major bump after a previous release on non-release"""
        self._assert_tag_version_local(
            spec=self._local_beta,
            tags=[('0.7.56', 'ci_base_build 106')],
            change_details="this won't cause a #major bump",
            expected_version=Version((0, 7, 2, 61, RevisionType.Beta)))

    def test_tag_version_local_release_major_bump(self):
        """Verifies handling of simulated major bump after a previous release on release"""
        self._assert_tag_version_local(
            spec=self._make_release_spec_for_build(111),
            tags=[('0.7.56', 'ci_base_build 106')],
            change_details="this will cause a #major bump",
            expected_version=Version((1, 0, 4, 0, RevisionType.Release)))

    def test_tag_version_local_rc_disabled_bump_on_merge_request(self):
        """Verifies handling of simulated minor after major bump after a previous release
        on non-release
        """
        self._assert_tag_version_local(
            spec=self._change_request_rc,
            tags=[('1.2.34', 'ci_base_build 1198')],
            change_details="this won't cause a #minor nor a #major bump",
            expected_version=Version((1, 2, 3, 70, RevisionType.RC)))

    def test_tag_version_local_rc_disabled_bump_on_master(self):
        """Verifies handling of simulated minor after major bump after a previous release
        on non-release
        """
        self._assert_tag_version_local(
            spec=self._master_rc,
            tags=[('1.2.34', 'ci_base_build 1198')],
            change_details="this won't cause a #minor nor a #major bump",
            expected_version=Version((1, 2, 3, 70, RevisionType.RC)))

    def test_tag_version_local_release_minor_and_major_bumps(self):
        """Verifies handling of simulated minor after major bump after a previous release
        on release where only the major bump should be used
        """
        self._assert_tag_version_local(
            spec=self._make_release_spec_for_build('1234'),
            tags=[('1.2.34', 'ci_base_build 1198')],
            change_details="this won't cause a #minor but a #major bump",
            expected_version=Version((2, 0, 4, 0, RevisionType.Release)))

    def test_apply_non_release(self):
        """Verifies trying to apply a non-release version raises an error"""
        target, repo = self._make_target(
            spec=self._master_rc,
            tags=[('1.2.34', 'ci_base_build 1198')],
            change_details="this won't cause a #minor nor a #major bump")
        with self.assertRaises(ValueError):
            target.apply()
        self.assertEqual([], repo.added, 'non-release tag added')

    def test_apply_release(self):
        """Verifies applying a release version"""
        target, repo = self._make_target(
            spec=self._make_release_spec_for_build('1234'),
            tags=[('1.2.34', 'ci_base_build 1198')],
            change_details="this won't cause a #minor but a #major bump")
        target.apply()
        self.assertEqual([('2.0.0', 'ci_base_build 1234')], repo.added,
                         'wrong/no release tag added')

    @staticmethod
    def _make_release_spec_for_build(build):
        return {'build version': build,
                'branch name': 'master',
                'merge request title': 'None',
                'job name': 'some_job'}

    def _assert_tag_version_local(self, spec, tags, change_details, expected_version):
        target, _ = self._make_target(spec, tags, change_details)
        self._assert_version_pattern(target)
        self.assertEqual(expected_version, target,
                         f'version not as expected: should be {expected_version}, but is {target}')

    def _make_target(self, spec, tags, change_details):
        change = Change()
        change.append_details(change_details)
        repo = RepoTestbed(tags, change)
        target = TagVersion(repo, CiIo(self._echo, spec))
        return target, repo


if __name__ == '__main__':
    unittest.main()
